import asyncio
import json
from pathlib import Path
import time
from typing import Optional, Union
import urllib.parse
import uuid

import httpx


class SyftError(Exception):
    pass


CONSTANTS = {
    "STORAGE_KEY": "syftbox-requests",
    "DEFAULT_SERVER_URL": "https://syftbox.net/",
    "DEFAULT_POLLING_INTERVAL": 3000,
    "DEFAULT_MAX_POLL_ATTEMPTS": 20,
    "DEFAULT_TIMEOUT": 5000,
    "MAX_BACKOFF_DELAY": 30000,
}

ANONYMOUS_EMAIL = "guest@syft.org"


class SyftRequest:
    def __init__(self, id: str, request_data: dict):
        self.id = id
        self.request_data = request_data
        self.status = "PENDING"
        self.timestamp = time.time()
        self.callbacks = []
        self.response_data = None
        self.error = None
        self.poll_timer = None
        self.poll_attempt = 0
        self.max_poll_attempts = CONSTANTS["DEFAULT_MAX_POLL_ATTEMPTS"]

    def update_status(self, status: str, data: dict):
        self.status = status
        if status == "SUCCESS":
            self.response_data = data
            self.poll_attempt = 0
        elif status == "ERROR":
            self.error = data
            self.poll_attempt = 0
        elif status == "POLLING":
            if not self.id or data:
                self.id = data

    def update_polling_progress(self, attempt: int, max_attempts: int):
        self.poll_attempt = attempt
        self.max_poll_attempts = max_attempts
        self.timestamp = time.time()


class PollingManager:
    def __init__(self, config: dict):
        self.interval = config.get("pollingInterval", 1000)
        self.max_attempts = config.get("maxPollAttempts", 30)

    async def poll(self, poll_fn, on_progress=None):
        attempt = 0
        stop = False
        while attempt < self.max_attempts and not stop:
            try:
                response = await poll_fn()
                if response.status_code == 200:
                    data = response.json()
                    if data.get("status") != "pending":
                        return response
                elif (
                    response.status_code == 500
                    and response.json().get("error")
                    == "No response exists. Polling timed out"
                ):
                    pass
                elif response.status_code == 404:
                    return response
                else:
                    return response

                if not stop:
                    await self._delay(self._get_backoff_delay(attempt))
                attempt += 1
                if on_progress:
                    on_progress(attempt, self.max_attempts)
            except Exception as error:
                if attempt == self.max_attempts - 1:
                    raise error

        raise SyftError("Polling timeout", "POLLING_TIMEOUT")

    def _get_backoff_delay(self, attempt: int) -> int:
        return min(1000 * (2**attempt), 30000)  # Max 30 seconds

    async def _delay(self, milliseconds: int):
        import asyncio

        await asyncio.sleep(milliseconds / 1000)


class SyftBoxSDK:
    def __init__(self, config: Optional[dict] = None):
        if config is None:
            config = {}
        self.config = config
        self.server_url = config.get("serverUrl", CONSTANTS["DEFAULT_SERVER_URL"])
        self.polling_manager = PollingManager(config)

    async def syft_fetch(self, syft_url: str, options: Optional[dict] = None):
        httpx_res, id = await self.syft_make_request(syft_url, options)
        return self.parse_httpx_res(httpx_res, id)

    async def syft_make_request(
        self,
        syft_url: str,
        body: Optional[dict] = None,
        from_email: str = ANONYMOUS_EMAIL,
        headers: Optional[dict] = None,
        method: str = "POST",
    ):
        if headers is None:
            headers = {}
        if body is None:
            body = {}

        headers["x-syft-from"] = from_email
        headers["x-syft-method"] = method
        headers["x-syft-body"] = json.dumps(body).encode("utf-8")

        request_data = {
            "syftUrl": syft_url,
            "fromEmail": from_email,
            "method": method,
            "headers": headers,
            "body": body,
        }
        id = str(uuid.uuid4())
        request = SyftRequest(id, request_data)

        httpx_res = await self.send_request(request)
        return httpx_res, id

    def parse_httpx_res(self, response: httpx.Response, request_id: str) -> dict:
        if not response.is_success:
            try:
                body = response.json()
            except:
                body = {}

            if (
                response.status_code == 500
                and body.get("error") == "No response exists. Polling timed out"
            ):
                return None
            elif response.status_code == 404:
                return {
                    "status": "ERROR",
                    "message": body.get("message", "No request found."),
                    "request_id": request_id,
                }
            else:
                raise SyftError(
                    f"Polling failed: {response.status_code}", "POLLING_ERROR"
                )
        return {
            "status_code": response.status_code,
            "headers": response.headers,
            "body": response.json(),
        }

    async def send_request(self, request: SyftRequest):
        syft_url = request.request_data["syftUrl"]
        from_email = request.request_data["fromEmail"]
        method = request.request_data["method"]
        headers = request.request_data["headers"]
        body = request.request_data["body"]

        combined_headers = {
            "Content-Type": "application/json",
            "x-syft-from": from_email,
            "timeout": str(self.config.get("timeout", CONSTANTS["DEFAULT_TIMEOUT"])),
            **headers,
        }

        try:
            raw_param = (
                f"&x-syft-raw={headers.get('x-syft-raw', '')}"
                if headers.get("x-syft-raw")
                else ""
            )
            msg_url = (
                f"{self.server_url}api/v1/send/msg?"
                f"x-syft-from={from_email}&"
                f"x-syft-url={urllib.parse.quote(syft_url)}"
                f"{raw_param}"
            )

            async with httpx.AsyncClient() as client:
                response = await client.request(
                    method=method,
                    url=msg_url,
                    headers=combined_headers,
                    content=json.dumps(body).encode("utf-8"),
                )
            if response.status_code == 202:
                response_body = response.json()
                print("initial response_body", response_body)
                if not response_body.get("request_id"):
                    raise SyftError(
                        "Accepted but missing request_id", "INVALID_RESPONSE"
                    )

                request.update_status("POLLING", response_body["request_id"])
                # for key, value in response.headers.items():
                #     print(f"Header: {key} = {value}")

                poll_url = response_body.get("data", {}).get("poll_url")
                location_header = response.headers.get("Location")

                poll_result = await self.poll_for_response(
                    {
                        "requestId": response_body["request_id"],
                        "pollUrlPath": poll_url or location_header,
                        "request": request,
                    }
                )

                request.update_status("SUCCESS", poll_result)
                return poll_result

            if response.is_success:
                response_data = response.json()
                request.update_status("SUCCESS", response_data)
                return response

            return response
        except Exception as error:
            request.update_status("ERROR", str(error))
            raise error

    async def poll_for_response(self, poll_data):
        """Poll for response from the server."""
        request_id = poll_data["requestId"]
        poll_url_path = poll_data["pollUrlPath"]
        request = poll_data["request"]

        poll_url = f"{self.server_url}{poll_url_path.lstrip('/')}"

        return await self.polling_manager.poll(
            poll_fn=lambda: self._poll_request(poll_url, request_id),
            on_progress=lambda attempt, max_attempts: request.update_polling_progress(
                attempt, max_attempts
            ),
        )

    async def _poll_request(self, poll_url, request_id):
        """Make a single poll request."""
        print("poll url", poll_url)
        async with httpx.AsyncClient() as client:
            response = await client.get(
                poll_url,
                headers={
                    "Content-Type": "application/json",
                    "Accept": "application/json",
                },
            )

        return response


class DirectSyftboxTransport(httpx.BaseTransport):
    def __init__(
        self, app_owner: str, app_name: str, sender_email: str = ANONYMOUS_EMAIL
    ) -> None:
        self.app_owner = app_owner
        self.app_name = app_name
        self.sender_email = sender_email

    @classmethod
    def from_config(cls, config_path: Path):
        pass

    def handle_request(self, request: httpx.Request) -> httpx.Response:
        sdk = SyftBoxSDK()
        if request.headers is None:
            headers = {}
        else:
            headers = dict(request.headers)

        headers.pop("content-length", None)

        body = json.loads(request.content) if request.content else {}

        response, _ = asyncio.run(
            sdk.syft_make_request(
                f"syft://{self.app_owner}/app_data/{self.app_name}/rpc/{request.url.path}",
                body=body,
                headers=headers,
                from_email=self.sender_email,
            )
        )

        response_headers = dict(response.headers)
        response_headers.pop("content-encoding", None)

        outer_response_body = (
            json.loads(response.content.decode("utf-8"))
            .get("data", {})
            .get("message", {})
        )
        response_content = json.dumps(outer_response_body.get("body", {})).encode(
            "utf-8"
        )

        syft_status_code = outer_response_body["status_code"]
        syft_headers = outer_response_body["headers"]

        http_response = httpx.Response(
            request=request,
            status_code=syft_status_code,
            headers=syft_headers,
            content=response_content,
        )

        return http_response

    def close(self) -> None:
        pass



from pathlib import Path
from typing import Optional

from fastapi import Request
from fastapi.responses import HTMLResponse
from pydantic import BaseModel

from fastsyftbox import FastSyftBox

app_name = Path(__file__).resolve().parent.name

app = FastSyftBox(
    app_name=app_name,
    syftbox_endpoint_tags=[
        "syftbox"
    ],  # endpoints with this tag are also available via Syft RPC
    include_syft_openapi=True,  # Create OpenAPI endpoints for syft-rpc routes
)


@app.get("/", response_class=HTMLResponse)
def root():
    return HTMLResponse(
        content=f"<html><body><h1>Welcome to {app_name}</h1>"
        + f"{app.get_debug_urls()}"
        + "</body></html>"
    )


class MessageModel(BaseModel):
    message: str
    name: Optional[str] = None


# tags=syftbox means also available via Syft RPC
# syft://{datasite}/app_data/{app_name}/rpc/endpoint
@app.api_route(
    path="/hello", methods=["GET", "POST", "PUT", "DELETE"], tags=["syftbox"]
)
async def hello_handler(request: Request) -> MessageModel:
    try:
        print("> Request Method: ", request.method)
        print("> Request Body: ", await request.body())
        if request.query_params:
            print("> Query Params: ", request.query_params)
        print("> SyftBox URL: ", request.state.syftbox_url)

        if request.method == "POST":
            json_data = await request.json()
            msg = MessageModel(**json_data)
            name = msg.name
        else:
            name = request.query_params.get("name")

        response = MessageModel(message=f"Hi {name}", name="Bob")
        return response
    except Exception as e:
        print("Error", e)
        raise e


# Debug your Syft RPC endpoints in the browser
app.enable_debug_tool(
    endpoint="/hello",
    example_request=str(MessageModel(message="Hello!", name="Alice").model_dump_json()),
    publish=True,
)



from pathlib import Path
from typing import Union

import httpx
from httpx import BaseTransport, Request, Response
from syft_core import Client as SyftBoxClient
from syft_core import SyftBoxURL, SyftClientConfig
from syft_rpc import rpc


def _read_content(obj: Union[Request, Response]) -> bytes:
    """Read content from a request or response object safely."""
    try:
        return obj.content
    except (httpx.RequestNotRead, httpx.ResponseNotRead):
        return obj.read()


class SyftFileSystemTransport(BaseTransport):
    def __init__(
        self,
        app_owner: str,
        app_name: str,
        data_dir: Path,
        sender_email: str = "guest@syftbox.com",
    ) -> None:
        self.app_owner = app_owner
        self.app_name = app_name
        self.data_dir = data_dir
        self.sender_email = sender_email

        self.app_dir = self.data_dir / app_owner / "app_data" / app_name
        self.rpc_dir = self.app_dir / "rpc"

    @classmethod
    def from_config(cls, config_path: Path) -> "SyftFileSystemTransport":
        pass

    def handle_request(self, request: Request) -> Response:
        body = _read_content(request)

        config = SyftClientConfig(
            client_url=8002,  # does not matter
            path="",  # does not matter
            data_dir=self.data_dir,
            email=self.sender_email,
        )

        syftbox_client = SyftBoxClient(
            conf=config,
        )
        expiry = "5s"

        syft_url = SyftBoxURL(
            f"syft://{self.app_owner}/app_data/{self.app_name}/rpc/{request.url.path}"
        )

        future = rpc.send(
            url=syft_url,
            method=request.method,
            body=body,
            headers=dict(request.headers.items()),
            cache=False,
            client=syftbox_client,
        )

        timeout_seconds = float(rpc.parse_duration(expiry).seconds)
        response = future.wait(timeout=timeout_seconds)
        http_response = httpx.Response(
            status_code=response.status_code.value,
            headers=response.headers,
            content=response.body,
        )

        return http_response

    def close(self) -> None:
        pass



from __future__ import annotations

import logging
from typing import Optional

import httpx
import uvicorn.logging
from syft_core import Client
from syft_event.server2 import SyftEvents
from syft_event.types import Request as SyftEventRequest
from syft_event.types import Response

handler = logging.StreamHandler()
formatter = uvicorn.logging.DefaultFormatter(fmt="%(levelprefix)s %(message)s")
handler.setFormatter(formatter)

logger = logging.getLogger("syftbox.http_bridge")
logger.setLevel(logging.INFO)
logger.propagate = False
logger.addHandler(handler)


from fastsyftbox.constants import SYFT_FROM_HEADER, SYFT_URL_HEADER

MAX_HTTP_TIMEOUT_SECONDS = 30


class SyftHTTPBridge:
    def __init__(
        self,
        app_name: str,
        http_client: httpx.AsyncClient,
        included_endpoints: list[str],
        syftbox_client: Optional[Client] = None,
    ):
        self.syft_events = SyftEvents(app_name, client=syftbox_client)
        self.included_endpoints = included_endpoints
        self.app_client = http_client  # Add the missing app_client attribute

    def start(self) -> None:
        self._register_rpc_handlers()
        self.syft_events.start()

    async def aclose(self) -> None:
        self.syft_events.stop()
        await self.app_client.aclose()

    def __enter__(self) -> SyftHTTPBridge:
        self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.aclose()

    async def _forward_to_http(
        self, request: SyftEventRequest, path: str
    ) -> httpx.Response:
        """Forward RPC request to HTTP endpoint."""
        method = self._get_method(request)
        headers = self._prepare_headers(request)
        response = await self.app_client.request(
            method=method,
            url=path,
            content=request.body,
            headers=headers,
            params=request.url.query or None,
        )
        return response

    def _get_method(self, request: SyftEventRequest) -> str:
        """Extract HTTP method from request."""
        try:
            return str(request.method) if request.method else "POST"
        except Exception as e:
            print(f"Error getting method Defaulting to POST: {e}")
            return "POST"

    def _prepare_headers(self, request: SyftEventRequest) -> dict:
        """Prepare headers for HTTP request."""
        headers = request.headers or {}
        headers[SYFT_URL_HEADER] = str(request.url)
        sender = getattr(request, "sender", None)
        if sender:
            headers[SYFT_FROM_HEADER] = str(sender)
        return headers

    def _register_rpc_handlers(self) -> None:
        for endpoint in self.included_endpoints:
            self._register_rpc_for_endpoint(endpoint)

    def _register_rpc_for_endpoint(self, endpoint: str) -> None:
        @self.syft_events.on_request(endpoint)
        async def rpc_handler(request: SyftEventRequest) -> Response:
            http_response = await self._forward_to_http(request, endpoint)
            return Response(
                body=http_response.content,
                status_code=http_response.status_code,
                headers=dict(http_response.headers),
            )



from pathlib import Path

import httpx

from fastsyftbox.direct_http_transport import ANONYMOUS_EMAIL, DirectSyftboxTransport
from fastsyftbox.transport import SyftFileSystemTransport

DEV_DEFAULT_OWNER_EMAIL = "guest@syftbox.com"


def default_dev_data_dir(app_name: str) -> Path:
    return Path(f"/tmp/{app_name}")


class SimpleRPCClient(httpx.Client):
    def __init__(
        self,
        *args,
        app_owner=None,
        app_name=None,
        data_dir=None,
        sender_email=ANONYMOUS_EMAIL,
        use_local_transport=False,
        **kwargs,
    ):
        self.use_local_transport = use_local_transport
        self.app_owner = app_owner
        self.app_name = app_name
        self.data_dir = data_dir
        self.sender_email = sender_email

        if use_local_transport:
            if self.app_owner is None:
                self.app_owner = DEV_DEFAULT_OWNER_EMAIL

            if self.data_dir is None:
                if self.app_name is None:
                    raise ValueError("data_dir or app_name must be provided")
                else:
                    self.data_dir = default_dev_data_dir(self.app_name)

            self.data_dir = Path(self.data_dir)
            transport = SyftFileSystemTransport(
                app_name=self.app_name,
                app_owner=self.app_owner,
                data_dir=self.data_dir,
                sender_email=self.sender_email,
            )

        else:
            if app_owner is None or app_name is None:
                raise ValueError("app_owner and app_name must be provided")

            transport = DirectSyftboxTransport(
                app_owner=self.app_owner,
                app_name=self.app_name,
                sender_email=self.sender_email,
            )

        super().__init__(
            *args, transport=transport, base_url="syft://localhost", **kwargs
        )

    @classmethod
    def for_syftbox_transport(
        cls, app_owner, app_name, sender_email=ANONYMOUS_EMAIL, **kwargs
    ):
        return cls(
            app_owner=app_owner,
            app_name=app_name,
            use_local_transport=False,
            sender_email=sender_email,
            **kwargs,
        )

    @classmethod
    def for_local_transport(
        cls,
        app_owner=None,
        app_name=None,
        data_dir=None,
        sender_email=ANONYMOUS_EMAIL,
        **kwargs,
    ):
        return cls(
            app_owner=app_owner,
            app_name=app_name,
            data_dir=data_dir,
            use_local_transport=True,
            sender_email=sender_email,
            **kwargs,
        )
